﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;

/// <summary>
/// This class creates a BoxCut of a geometry object
/// </summary>
public class BoxCut : Cut
{

	private Transform geoTrans;
	private int[] triangles;
	private Mesh startMesh = null;
	private ParameterSpace vv;
	private ParameterSpace vv2;

	/// <summary>
	/// Constructor of bux cut
	/// </summary>
	/// <param name="geoTrans">The geometry object which should be cuted</param>
	public BoxCut(Transform geoTrans){
		this.geoTrans = geoTrans;
	}

	/// <summary>
	/// Precomputes the box Cut for the given target. The bounds are calculated to cut the geometry object iin such a 
	/// way that the target is seen from the camera. The created Bounds are calculated with the ParameterSpace
	/// </summary>
	/// <param name="camera">The transform of the camera</param>
	/// <param name="target">The transform of the target</param>
	public void precompute(Transform camera, List<Transform> targets)
	{
	 	geoTrans.gameObject.SetActive(true);
		if (startMesh == null)
			startMesh = geoTrans.GetComponent<MeshFilter>().mesh;
		triangles = geoTrans.GetComponent<MeshFilter>().mesh.triangles;
		//var active = target.gameObject.activeInHierarchy;
		foreach(var target in targets) target.gameObject.SetActive(true);
		var geoBounds = geoTrans.GetComponent<MeshCollider> ().bounds;
		var parameters = new List<ParameterSpace>();
        Bounds combinedBounds = targets[0].GetComponent<MeshCollider>().bounds;
        for(var i = 1; i < targets.Count; i++)
        {
            combinedBounds.Encapsulate(targets[i].GetComponent<MeshCollider>().bounds);
        }
        //var bbTarget = target.GetComponent<MeshCollider>().bounds;
        var bbTarget = combinedBounds;
        drawBoundingBox(bbTarget,Color.yellow);
		drawBoundingBox(geoBounds,Color.green);
		for (var i = -2; i <= 2; i++)
		{
			for (var j = -2; j <= 2; j++)
			{
				for (var k = -2; k <= 2; k++)
				{
					for (var ii = -2; ii <= 2; ii++)
					{
						for (var jj = -2; jj <= 2; jj++)
						{
							for (var kk = -2; kk <= 2; kk++)
							{
								if(i == ii && k == kk && j == jj)
									continue;
								if(i == 0 || j == 0 || k == 0 || ii == 0 || jj == 0 || kk == 0)
									continue;
								for (var steps = 0.01f;  steps < 1f; steps += 0.01f)
								{
									Vector3 destination = bbTarget.center;
									Vector3 mult = new Vector3(i,j,k);
									var v = new Vector3(ii, jj, kk);
									mult.x = bbTarget.extents.x/mult.x;
									mult.y = bbTarget.extents.y/mult.y;
									mult.z = bbTarget.extents.z/mult.z;
									v.x = bbTarget.extents.x/v.x;
									v.y = bbTarget.extents.y/v.y;
									v.z = bbTarget.extents.z/v.z;
									mult = mult * steps + v * (1f - steps);
									destination += mult;
									var ray = new Ray(camera.position,destination-camera.position);
									RaycastHit[] hits = Physics.RaycastAll(ray);
									if (hits.Any(h => h.transform.name == geoTrans.name))
									{
										var hit = hits.Where(h => h.transform.name == geoTrans.name).First();
										var par = new ParameterSpace(hit.point, geoBounds);
                                        //var camToTargetMag = (target.position - camera.position).magnitude;
                                        var camToTargetMag = (getCenterOfTarget(targets) - camera.position).magnitude;
                                        var camToHittMag = (hit.point - camera.position).magnitude;
										if(camToTargetMag > camToHittMag)
											parameters.Add(par);
									}
								}
							}
						}
					}
					
				}
			}
		}
		vv = ParameterSpace.Largest(parameters);
		vv2 = ParameterSpace.Smallest(parameters);
		if (vv == null || vv2 == null)
		{
			return;
		}
		var bounds = ParameterSpace.CreateNewBounds(vv.GetPosition(), vv2.GetPosition());
		drawBoundingBox(bounds,Color.red);
        //foreach (var target in targets) target.gameObject.SetActive(active);
        
	}


    /// <summary>
    /// All the triangles which lay inside the already calculated u, v, w, u', v' and w' values are removed from the mesh.
    /// </summary>
    /// <param name="target">The transform of the target geometry</param>
    /// <param name="occlusionLevel">The current level in the occlusion tree</param>
    /// <param name="camera">The current camera</param>
    public void computeFinalCut(List<Transform> targets, int occlusionLevel, Camera camera)
    {
		geoTrans.gameObject.SetActive(true);
		if (vv == null || vv2 == null)
			return;
        var bounds = ParameterSpace.CreateNewBounds(vv.GetPosition(), vv2.GetPosition());
		var mesh = geoTrans.GetComponent<MeshFilter>().mesh; 
		Vector3[] vertices = mesh.vertices;
		Vector2[] uvs = mesh.uv;
	    Vector3[] normals = mesh.normals;
		mesh.Clear();
		mesh.vertices = vertices;
		mesh.uv = uvs;
		mesh.normals = normals;
		var newTriangles = new List<int>();
		bounds = new Bounds(bounds.center,bounds.size + bounds.size*(occlusionLevel+4)/20f);
		drawBoundingBox(bounds,Color.magenta);
		for (int i = 0; i < triangles.Length-2; i += 3)
		{
			var a = geoTrans.TransformPoint(vertices[triangles[i]]);
			var b = geoTrans.TransformPoint(vertices[triangles[i+1]]);
			var c = geoTrans.TransformPoint(vertices[triangles[i+2]]);
			var triangle = new List<Vector3>();
			triangle.Add(a);
			triangle.Add(b);
			triangle.Add(c);
			int hits = 0;
			foreach (var vert in triangle)
			{
				if (bounds.Contains(vert))
					hits++;
			}
			if(hits >= 1)
				continue;
			newTriangles.Add(triangles[i]);
			newTriangles.Add(triangles[i+1]);
			newTriangles.Add(triangles[i+2]);
		}
		mesh.triangles = newTriangles.ToArray();
		geoTrans.GetComponent<MeshCollider> ().sharedMesh = mesh;
    }

	/// <summary>
	/// Resets the geometry objects triangles to initial value
	/// </summary>
	public void resetGeometry()
	{
		if(startMesh == null) return;
		geoTrans.GetComponent<MeshCollider>().sharedMesh = startMesh;
	}

	/// <summary>
	/// This class calculates the position of a point inside a given bound by parameterizing it.
	/// </summary>
	private class ParameterSpace{
		private float u;
		private float v;
		private float w;
		private Bounds _bounds;

		public ParameterSpace(float u, float v, float w, Bounds bounds){
			this.u = u;
			this.v = v;
			this.w = w;
			_bounds = bounds;
		}
		
		public ParameterSpace(Vector3 point, Bounds bounds){
			_bounds = bounds;
			var newPos = point - _bounds.min;
			this.u = Mathf.Abs(newPos.x / (_bounds.size.x));
			this.v = Mathf.Abs(newPos.y / (_bounds.size.y));
			this.w = Mathf.Abs(newPos.z / (_bounds.size.z));
			
		}
		
		/// <summary>
		/// Returns a new ParameterSpace object with the maximum u, v and w of each objet in the list
		/// </summary>
		/// <param name="parameterSpaces">A list of ParameterSpace objects</param>
		/// <returns>A new object with the largest parameters u,v,w from the input list</returns>
		public static ParameterSpace Largest(List<ParameterSpace> parameterSpaces)
		{
			ParameterSpace ret = null;
			if (parameterSpaces.Count == 0)
				return null;
			var u = 0f;
			var v = 0f;
			var w = 0f;
			foreach (var par in parameterSpaces)
			{
				if (par.u > u)
					u = par.u;
				if (par.v > v)
					v = par.v;
				if (par.w > w)
					w = par.w;
			}
			ret = new ParameterSpace(u,v,w,parameterSpaces.First()._bounds);
			return ret;
		}
		
		/// <summary>
		/// Returns a new ParameterSpace object with the minimum u, v and w of each objet in the list
		/// </summary>
		/// <param name="parameterSpaces">A list of ParameterSpace objects</param>
		/// <returns>A new object with the smallest parameters u,v,w from the input list</returns>
		public static ParameterSpace Smallest(List<ParameterSpace> parameterSpaces)
		{
			ParameterSpace ret = null;
			if (parameterSpaces.Count == 0)
				return null;
			var u = 1f;
			var v = 1f;
			var w = 1f;
			foreach (var par in parameterSpaces)
			{
				if (par.u < u)
					u = par.u;
				if (par.v < v)
					v = par.v;
				if (par.w < w)
					w = par.w;
			}
			ret = new ParameterSpace(u,v,w,parameterSpaces.First()._bounds);
			return ret;
		}

		/// <summary>
		/// Returns the position of the parameterized object
		/// </summary>
		/// <returns>The non parameterized position</returns>
		public Vector3 GetPosition()
		{
			var newPos = _bounds.min;
			var ext = _bounds.size;
			ext.x *= u;
			ext.y *= v;
			ext.z *= w;
			newPos += ext;
			return newPos;
		}
		
		/// <summary>
		/// Creates a new Bounds element which lies in the midlle of the two parameters and has one of the vertices on a parameter.
		/// </summary>
		/// <param name="par1">The first Parameter</param>
		/// <param name="par2">The second Parameter</param>
		/// <returns>The new bounds</returns>
		public static Bounds CreateNewBounds(Vector3 par1, Vector3 par2)
		{
			Vector3 newCenter = (par1+par2)/2f;
			Vector3 newExtents = newCenter - par1;
			newExtents.x = Mathf.Abs(newExtents.x);
			newExtents.y = Mathf.Abs(newExtents.y);
			newExtents.z = Mathf.Abs(newExtents.y);
			newExtents *= 2f;
			return new Bounds(newCenter,newExtents);
		}
		
		/// <summary>
		/// The ToString method of the ParameterSpace class
		/// </summary>
		/// <returns>A strinf representing the ParameterSpace object</returns>
		public string ToString()
		{
			return "u: " + u + " v: " + v + " w: " + w + " center: " + _bounds.center + " size: " + _bounds.size;
		}
	}
	
	/// <summary>
	/// Returns the transform of the geometry object
	/// </summary>
	/// <returns>The transform of the geometry object</returns>
	public Transform getTransform(){
		return geoTrans;
	}
	
	/// <summary>
	/// This function draws a line between each vertices of the boundingbox, with the given color
	/// </summary>
	/// <param name="bounds">bounds the boundingbox</param>
	/// <param name="color">color the color for the lines</param>
	private void drawBoundingBox(Bounds bounds, Color color){
		var debugL = new List<Vector3>();
		for (var i = -1; i <= 1; i += 2)
		{
			for (var j = -1; j <= 1; j += 2)
			{
				for (var k = -1; k <= 1; k += 2)
				{
					Vector3 destination = bounds.center;
					Vector3 mult = new Vector3(i,j,k);
					mult.x *= bounds.extents.x;
					mult.y *= bounds.extents.y;
					mult.z *= bounds.extents.z;
					destination += mult;
					debugL.Add(destination);
					
				}
			}
		}
		drawDebugLines(debugL, color);
	}

	/// <summary>
	/// Computes the center in world coordinates of a geometry
	/// </summary>
	/// <param name="targets">The transform of a geometry</param>
	/// <returns>Center of the geometry</returns>
    private Vector3 getCenterOfTarget(List<Transform> targets)
    {
        Vector3 center = Vector3.zero;
        foreach (Transform target in targets)
        {
            Mesh mesh = target.GetComponent<MeshFilter>().mesh;
            center += target.TransformPoint(mesh.bounds.center);
        }
        return center / targets.Count;
    }

	/// <summary>
	/// This function draws a line between each polygon of the list, with the given color
	/// </summary>
	/// <param name="polygon">polygon all the point to draw lines to</param>
	/// <param name="color">color the color for the lines</param>
	private void drawDebugLines(List<Vector3> polygon, Color color){
		foreach (var v1 in polygon)
		{
			foreach (var v2 in polygon)
			{
				Debug.DrawLine (v1, v2, color, 10000, false);
			}
		}
	}
}

