﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;

/// <summary>
/// Freeform window cuts are defined by a  bounding curve and a single parameter u that represents the size of the window.
/// At u = 0, the window is fully closed. At u = 1, the window is fully open. To accelerate the computation of this mapping,
/// the system precomputes bounding curves at uniformly spaced u values (0.1 steps).
/// </summary>
public class FreeformWindowCut : Cut {
	
	private Dictionary<int, List<Vector3>> preWindowBoundaries; //The boundary for a given u
	private Dictionary<int, List<int>> preTriangles;	//The triangles to Render for a given u

	private List<Vector3> border;	//A list of sorted vectors that specify the border of an element
	private Vector3 targetCenter;

    private Transform geoTrans;


	/// <summary>
	/// The constructor of the <c>FreeformWindowCut</c> class
	/// </summary>
	/// <param name="geoTrans">The transform of the geometry that is associated with this cut.</param>
    public FreeformWindowCut(Transform geoTrans){
		this.geoTrans = geoTrans;
		border = new List<Vector3> ();
	}

	
	/// <summary>
	/// Precomputes the window Cut. First, it computes the border for all possible u (0, 0.1 , ... , 1) and saves then the cutted meshes for each u.
	/// </summary>
	/// <param name="camera">The transform of the camera</param>
	/// <param name="targets">The transform of the targets</param>
	public void precompute(Transform camera, List<Transform> targets)
	{
        preTriangles = new Dictionary<int, List<int>>();
        border = findBorder (camera.GetComponent<Camera> ());
		if (border.Count > 0) {
			targetCenter = getCenterOfTarget (targets);
			preWindowBoundaries = computePredefinedBoundaries (border, targetCenter);
            preTriangles.Add(0, geoTrans.GetComponent<MeshFilter>().mesh.triangles.ToList());
            preTriangles.Add(10, new List<int>());
            
            //computeDefaultCuts (camera.GetComponent<Camera>());
        }
	}

	/// <summary>
	/// The function detects the outer border of the geometry on the camera screen.
	/// This is done by transforming each vertex to the screen and then computing the contour of the screen points.
	/// </summary>
	/// <param name="camera">The camera that is used to compute the screen space</param>
	/// <returns></returns>
	private List<Vector3> findBorder(Camera camera){
		Debug.Log ("Start find border of Geoemtry " +geoTrans.name );
		List<Vector3> borderVertices = new List<Vector3> ();

		//Get all vertices that occlude other vertices
		Vector3[] vertices = geoTrans.GetComponent<MeshFilter>().mesh.vertices;
		int[] triangles = geoTrans.GetComponent<MeshFilter> ().mesh.triangles;
		List<int> alreadyProcessed = new List<int> ();
		List<Vector3> vFacingCam = new List<Vector3> ();
		for (int i = 0; i < vertices.Length; i++) {
			Vector3 vertexOnScreen = camera.WorldToScreenPoint(geoTrans.TransformPoint (vertices[i]));
			vertexOnScreen = new Vector3 (Mathf.Round(vertexOnScreen.x), Mathf.Round(vertexOnScreen.y), vertexOnScreen.z);
			if (!vFacingCam.Any (v => v.x == vertexOnScreen.x && v.y == vertexOnScreen.y)) {
				vFacingCam.Add (vertexOnScreen ); 
			}
		}
		List<Vector3> contour = computeContourPoints (vFacingCam);
		List<Vector3> contourInWorld = new List<Vector3> ();
		for (int i = 0; i < contour.Count; i++) {
			contourInWorld.Add (camera.ScreenToWorldPoint(contour[i]));
			//Debug.DrawLine (camera.ScreenToWorldPoint(contour[i]), new Vector3(camera.ScreenToWorldPoint(contour[i]).x + 0.01f, camera.ScreenToWorldPoint(contour[i]).y + 0.01f, camera.ScreenToWorldPoint(contour[i]).z + 0.01f), Color.red, 10000, true);
		}
	
		return contourInWorld;
	}

	/// <summary>
	/// Returns the transform of the geometry object
	/// </summary>
	/// <returns>The transform of the geometry object</returns>
	public Transform getTransform(){
		return geoTrans;
	}

	
	/// <summary>
	/// Computes the center in world coordinates of a geometry
	/// </summary>
	/// <param name="target">The transform of a geometry</param>
	/// <returns>Center of the geometry</returns>
	private Vector3 getCenterOfTarget(List<Transform> targets){
       Vector3 center = Vector3.zero;
        foreach (Transform target in targets)
        {
            Mesh mesh = target.GetComponent<MeshFilter>().mesh;
            center += target.TransformPoint(mesh.bounds.center);
        }
		return center/targets.Count;
	}
	
	/// <summary>
	/// Computes the center point of a polygon
	/// </summary>
	/// <param name="polygonList">list of 3D vectors</param>
	/// <returns>The center of the polygon</returns>
	private Vector3 computeCentroid(List<Vector3> polygonList){
		Vector3 centroid = Vector3.zero;
		float minX = polygonList.Min (v => v.x);
		float maxX = polygonList.Max (v => v.x);
		float minY = polygonList.Min (v => v.y);
		float maxY = polygonList.Max (v => v.y);
		float minZ = polygonList.Min (v => v.z);
		float maxZ = polygonList.Max (v => v.z);

		return new Vector3 (minX + (maxX-minX)/2, minY + (maxY-minY)/2, minZ +(maxZ-minZ)/2);
		//Debug.DrawLine(centroid, centroid+new Vector3(0.01f, 0.01f, 0.01f), Color.red, 10000, false);
	}

	/// <summary>
	/// Compute the boundary for every u. This is done by moving the polygon points toward the centroid.
	/// </summary>
	/// <param name="outerBorder">The outerBorder of the cut - u = 1</param>
	/// <param name="centroid">Center of the boundaries</param>
	/// <returns>dictionary that contains a polygon list for each u</returns>
	private Dictionary<int, List<Vector3>> computePredefinedBoundaries(List<Vector3> outerBorder, Vector3 centroid){

		Dictionary<int, List<Vector3>> preWindowBoundaries = new Dictionary<int, List<Vector3>> ();
		preWindowBoundaries.Add (10, outerBorder);
		drawDebugLines (preWindowBoundaries [10], Color.blue);

		preWindowBoundaries.Add (9, computeBoundary(outerBorder, centroid, 0.1f));
		drawDebugLines (preWindowBoundaries [9], Color.blue);

		preWindowBoundaries.Add (8, computeBoundary(outerBorder, centroid, 0.2f));
		drawDebugLines (preWindowBoundaries [8], Color.blue);

		preWindowBoundaries.Add (7, computeBoundary(outerBorder, centroid, 0.3f));
		drawDebugLines (preWindowBoundaries [7], Color.blue);

		preWindowBoundaries.Add (6, computeBoundary(outerBorder, centroid, 0.4f));
		drawDebugLines (preWindowBoundaries [6], Color.blue);

		preWindowBoundaries.Add (5, computeBoundary(outerBorder, centroid, 0.5f));
		drawDebugLines (preWindowBoundaries [5], Color.blue);

		preWindowBoundaries.Add (4, computeBoundary(outerBorder, centroid, 0.6f));
		drawDebugLines (preWindowBoundaries [4], Color.blue);

		preWindowBoundaries.Add (3, computeBoundary(outerBorder, centroid, 0.7f));
		drawDebugLines (preWindowBoundaries [3], Color.blue);

		preWindowBoundaries.Add (2, computeBoundary(outerBorder, centroid, 0.8f));
		drawDebugLines (preWindowBoundaries [2], Color.blue);

		preWindowBoundaries.Add (1, computeBoundary(outerBorder, centroid, 0.9f));
		drawDebugLines (preWindowBoundaries [1], Color.blue);

		preWindowBoundaries.Add (0, new List<Vector3>());
		return preWindowBoundaries;

	}
	
	/// <summary>
	/// Computes the cutted mesh for each u.
	/// </summary>
	/// <param name="camera">camera The current viewpoint</param>
	private void computeDefaultCuts(Camera camera){
		preTriangles = new Dictionary<int, List<int>> ();
		List<int> us = new List<int>(preWindowBoundaries.Keys);
		for (int i = 0; i < us.Count; i++) {
			if (us [i] == 0) {	//Whole mesh is visible
				preTriangles.Add (us [i], geoTrans.GetComponent<MeshFilter> ().mesh.triangles.ToList());
			} else if (us [i] == 1) {
				preTriangles.Add (us [i], new List<int> ());
			} else {
				preTriangles.Add (us [i], computeCut (us[i], camera));
			}
		}
	}

	/// <summary>
	/// Renders the cutted mesh for a given u.
	/// </summary>
	/// <param name="u">Indicates the extend of the cut</param>
	private void cutMesh(int u){
        Mesh mesh = geoTrans.GetComponent<MeshFilter> ().mesh; 
		Vector3[] vertices = mesh.vertices;
		Vector2[] uvs = mesh.uv;
		Vector3[] normals = mesh.normals;
		mesh.Clear ();
		mesh.vertices = vertices;
		mesh.uv = uvs;
		mesh.normals = normals;
		mesh.triangles = preTriangles[u].ToArray ();
		Debug.Log("triangles: " + mesh.triangles.Length + " vertices: " + mesh.vertices.Length + " normals: " + mesh.normals.Length);
		geoTrans.GetComponent<MeshCollider> ().sharedMesh = mesh;
		
	}

	/// <summary>
	/// Detects all triangles that are visible when cutting the mesh with a given u.
	/// </summary>
	/// <param name="u">Extend of cut (Range 0, 0.1, ... ,1)</param>
	/// <param name="camera">The current camera</param>
	/// <returns>Returns a list with triangles that are visible in the cutted mesh.</returns>
	public List<int> computeCut(int u, Camera camera){
		List<int> visibleTriangles = new List<int> ();
		List<Vector3> borderPolygon = preWindowBoundaries[u];
		List<Vector3> borderOnScreen = new List<Vector3> ();
		for (int i = 0; i < borderPolygon.Count; i++) {
			borderOnScreen.Add (camera.WorldToScreenPoint(borderPolygon[i]));
		}
		if (borderPolygon.Count > 3) {
			Mesh mesh = geoTrans.GetComponent<MeshFilter> ().mesh;
			Vector3[] vertices = mesh.vertices;
			int[] triangles = mesh.triangles;

			for (int i = 0; i < triangles.Length-2; i = i +3) {
				int index1 = triangles [i];
				int index2 = triangles [(i + 1) ];
				int index3 = triangles [(i + 2) ];
				Vector3 vertex1 = vertices [index1];
				Vector3 vertex2 = vertices [index2];
				Vector3 vertex3 = vertices [index3];

			/*	Vector3 side1 = vertex2 - vertex1;
				Vector3 side2 = vertex3 - vertex1;

				Vector3 normalVector = Vector3.Cross (side1, side2);
				normalVector = normalVector / normalVector.magnitude;
				float dotProduct = Vector3.Dot (normalVector, camera.transform.forward);
				/*bool isInside = pointIsInsidePolygon (camera.WorldToScreenPoint(geoTrans.TransformPoint (vertex1)),borderOnScreen);
				if (isInside) {
					isInside = pointIsInsidePolygon (camera.WorldToScreenPoint(geoTrans.TransformPoint (vertex2)), borderOnScreen);
					if (isInside) {
						isInside = pointIsInsidePolygon (camera.WorldToScreenPoint(geoTrans.TransformPoint (vertex3)), borderOnScreen);
					}
				}*f/
				if (dotProduct <= 0) {*/
					Vector3 centroid = computeCentroid (new List<Vector3> (){ vertex1, vertex2, vertex3 });
					bool isInside = pointIsInsidePolygon (camera.WorldToScreenPoint (geoTrans.TransformPoint (centroid)), borderOnScreen);

                    if (u == 2 || u == 4|| u == 6 || u == 8) {
                    //if (true) { 
						if (isInside) {		
							//	if ((u*10)%2 == 0 && isInside || (u*10)%2 ==1 && !isInside) {	//Verstehe nicht warum man das hier so berechnen muss.
							visibleTriangles.Add (index1);
							visibleTriangles.Add (index2);
							visibleTriangles.Add (index3);
						} 
					} else {
						if (!isInside) {	
                        //	if ((u*10)%2 == 0 && isInside || (u*10)%2 ==1 && !isInside) {	//Verstehe nicht warum man das hier so berechnen muss.
							visibleTriangles.Add (index1);
							visibleTriangles.Add (index2);
							visibleTriangles.Add (index3);
						} 
					}
				/*} else {	
					visibleTriangles.Add (index1);
					visibleTriangles.Add (index2);
					visibleTriangles.Add (index3);
				}*/
			}
		}
		return visibleTriangles;
	}

	/// <summary>
	/// Detects the correct u values. It starts with u = 0 (window fully closed) and checks if the target geometry is 
	/// fully visible. If not, it increases the extend of the cut until the whole target is visible. In order that the 
	/// cutting volumns of the lower levels are still visible, the extend u is updated based on the current occlusionLevel.
	/// </summary>
	/// <param name="targets">The transform of the target geometries</param>
	/// <param name="occlusionLevel">The current level in the occlusion tree</param>
	/// <param name="camera"The current camera></param>
	public void computeFinalCut(List<Transform> targets, int occlusionLevel, Camera camera){
		//Disable all other objects in scene
		geoTrans.gameObject.SetActive(true);
		int u = 1;
		for (; u <= 10; u++) {
			Debug.Log ("Test with u: " + u);
            if (!preTriangles.ContainsKey(u))
            {
                preTriangles.Add(u, computeCut(u, camera));
            }
			cutMesh (u);
            
            if (targetIsFullyVisible(geoTrans, targets, camera))
            {
                break;
            }
		}
		u = u + occlusionLevel;
        
        if (u < 10)
        {
            if (!preTriangles.ContainsKey(u))
            {
                preTriangles.Add(u, computeCut(u, camera));
            }
            cutMesh(u);
        }
			
		geoTrans.gameObject.SetActive(false);
	}

	/// <summary>
	/// Resets the geometry objects triangles to initial value
	/// </summary>
	public void resetGeometry()
	{
		if(preTriangles != null && preTriangles.Any())
			cutMesh(0);
	}

	/// <summary>
	/// Checks if the target is fully visible from a viewpoint and a given cut model
	/// </summary>
	/// <param name="cutModel">The transform of the cut geometry</param>
	/// <param name="target"><The transform of the target/param>
	/// <param name="camera">The current camera that specifies the viewpoint</param>
	/// <returns>True, if the target is fully visible. False, if the target is occluded by the cutModel</returns>
	private bool targetIsFullyVisible(Transform cutModel, List<Transform> targets, Camera camera){
	//	Mesh mesh = target.GetComponent<MeshFilter> ().mesh;
//		Vector3[] vertices = mesh.vertices;
		//for (int i = 0; i < vertices.Length; i++) {
		//TODO Check if bounding box is visible
        Bounds combinedBounds = targets[0].GetComponent<MeshCollider>().bounds;
        for (var i = 1; i < targets.Count; i++)
        {
            combinedBounds.Encapsulate(targets[i].GetComponent<MeshCollider>().bounds);
        }
       for (var i = 0; i <= 1; i++)
        {
            for (var j = 0; j <= 1; j++)
            {
                for (var k = 0; k <= 1; k++)
                {
                   var extend = (new Vector3(i, 0, 0)) * combinedBounds.size.x + (new Vector3(0, j, 0)) * combinedBounds.size.y + (new Vector3(0, 0, k)) * combinedBounds.size.z;
                    Vector3 bbPoint = combinedBounds.center + extend;
                    Ray ray = camera.ViewportPointToRay(camera.WorldToViewportPoint(bbPoint));
                    RaycastHit[] hits = Physics.RaycastAll(ray);
                    foreach (RaycastHit hit in hits)
                    {
                        if (hit.transform.name.Equals(geoTrans.name))
                        {
                            return false;
                        }
                    }
                }
            }
        }
/*
        for (int i = 0; i < Screen.width; i++) {
			for (int j = 0; j < Screen.height; j++) {
				//Vector3 viewportVertex = camera.WorldToViewportPoint (target.TransformPoint (vertices [i]));
				Ray ray = camera.ScreenPointToRay(new Vector2(i,j));
				RaycastHit[] hits = Physics.RaycastAll (ray);
				bool targetHit = false;
				bool cutHit = false;
				foreach (RaycastHit hit in hits) {
					if (hit.transform.name.Equals (target.name)) {
						countHittingTarget++; 
						targetHit = true;
					} else if (hit.transform.name.Equals (geoTrans.name)) {
						cutHit = true;
					}
				}
				if (targetHit && cutHit) {
					countHittingBoth++;
					return false;
				}
				//}
			}
		}
		/*0Debug.Log ("Total Hit Target: " + countHittingTarget + "; Count hitting both: "+ countHittingBoth);
		if (countHittingBoth > 0) {
			return false;
		}*/
		return true;
	}
	
	/// <summary>
	/// Detects if a point is inside or outside a polygon. It considers only the x and y coordinate
	/// </summary>
	/// <param name="pointOnScreen">A point on screen</param>
	/// <param name="polygon">A polygon that contains point on screen</param>
	/// <returns>True, if the point is inside polygon and false if it is outside.</returns>
	private bool pointIsInsidePolygon(Vector3 pointOnScreen, List<Vector3> polygon){

		Vector2 ps1 = new Vector2(pointOnScreen.x, pointOnScreen.y);
		Vector2 pe1 = new Vector2 (pointOnScreen.x + 100000, pointOnScreen.y);
		int intersectionCount = 0;
		for (int i = 0; i < polygon.Count; i++) {
			Vector2 ps2 = new Vector2(polygon [i].x, polygon[i].y);
			Vector2 pe2 = new Vector2(polygon [(i+1)%polygon.Count].x, polygon[(i+1)%polygon.Count].y);
			Vector2 s = Vector2.zero;
			try{
				s = lineIntersectionPoint (ps1,pe1, ps2, pe2);
				if (s.x > pointOnScreen.x && (s.y < ps2.y && s.y > pe2.y || s.y < pe2.y && s.y > ps2.y)) {
					//Debug.Log ("Schnittpunkt ist zwischen den zwei Polygonpunkten");
					intersectionCount++;
				}
			}catch(System.Exception e){
			//	Debug.Log ("Lines are correctly parallel.");
				if (pointOnScreen.y == pe1.y) {
					intersectionCount++;
				}
			}
		}
		if (intersectionCount == 0) {
			//Debug.Log ("Vertex schneidet keine Border - Auserhalb");
			return false;
		} else if (intersectionCount % 2 == 1) {	
			//Debug.Log ("Vertex schneidet ungerade Anzahl an Linien - Innerhalb");
			return true;
		} else if (intersectionCount % 2 == 0) {
			//Debug.Log ("Vertex schneidet gerade anzahl an Linien - Auserhalb");
			return false;
		}
		return false;
	}

	
	/// <summary>
	/// This method computes a new boundary by shifting the points in the outerBoundary toward the centroid
	/// </summary>
	/// <param name="outerBoundary">The contour of the object in screen space</param>
	/// <param name="centroid">The center of the targetObject in screen space</param>
	/// <param name="percentDistance">The amount to shift in percent</param>
	/// <returns>The new boundarys</returns>
	private List<Vector3> computeBoundary(List<Vector3> outerBoundary, Vector3 centroid, float percentDistance){
		outerBoundary.Add (outerBoundary [0]);
		List<Vector3> newBoundary = new List<Vector3> ();
		for (int i = 0; i < outerBoundary.Count-1; i++) {
			Vector3 origDistance = centroid - outerBoundary[i];
			newBoundary.Add(outerBoundary[i] + origDistance*percentDistance);
		}	
		return newBoundary;
	}
	
	/// <summary>
	/// draws debug Lines
	/// </summary>
	/// <param name="polygon">The polygon to draw in the Scene View</param>
	/// <param name="color">The color of the polygon</param>
	private void drawDebugLines(List<Vector3> polygon, Color color){
		for (int i = 0; i < polygon.Count - 1; i++) {
			Debug.DrawLine (polygon[i], polygon[i + 1], color, 10000, false);
		}
		Debug.DrawLine (polygon [polygon.Count - 1], polygon [0], color, 10000, false);
	}
	
	/// <summary>
	/// Computes the Countour Points of a list of 2D Points and returns a list of the indices.
	/// The list of indices describes the order of the points in order to result in a contour.
	/// </summary>
	/// <param name="pointList">The point list in screen spave</param>
	/// <returns>The contour points of the point list</returns>
	public static List<Vector3> computeContourPoints(List<Vector3> pointList)
	{
		if (pointList.Count < 3) return pointList;
		List<int> alreadyProcessed = new List<int> ();
		List<Vector3> convexHull = new List<Vector3>();
		//First get startPoint
		float minY = pointList.Min(p => p.y);
		List<Vector3> startPos = pointList.Where(p => p.y == minY).ToList();
		if (startPos.Count > 1)
		{
			float minX = startPos.Min(p => p.x);
			startPos = startPos.Where(p => p.x == minX).ToList();
		}
		convexHull.Add(startPos[0]);
		//  pointList.Remove(startPos[0]);
		// int hullIndex = convexHull.Count - 1;
		int hullIndex = pointList.IndexOf(startPos[0]);

		bool convexHullFound = false;
		while (!convexHullFound)
		{
			bool convexHullPointFound = false;
			int refPos = 0;
			if (refPos == hullIndex)
				refPos = 1;
			
			while (!convexHullPointFound)
			{
				Vector3 line = (pointList[refPos] - pointList[hullIndex]).normalized;
				bool hasPointOnLeftSide = false;
				//TODO Check if a vector is on the left side of line
				for (int j = refPos; j < pointList.Count; j++)
				{
					if (j != refPos && j != hullIndex)
					{
						Vector3 refLine = (pointList[j] - pointList[hullIndex]).normalized;
						float angle = Mathf.Atan2(refLine.y, refLine.x) - Mathf.Atan2(line.y, line.x); //Ich glaube das ist richtig
						if (angle > Mathf.PI) angle -= 2 * Mathf.PI;
						else if (angle < -Mathf.PI) angle += 2 * Mathf.PI;
						if (angle >= 0)
						{
							if (!convexHull.Contains (pointList [j]) || convexHull[0] == pointList[j]) {
								//Point is on the left side
								hasPointOnLeftSide = true;
								refPos = j;
								break;
							}
						}
					}
				}
				if (!hasPointOnLeftSide)
				{
					convexHullPointFound = true;
					if (pointList[refPos] == startPos[0])
					{
						convexHullFound = true;
					}
					else
					{
						if (!alreadyProcessed.Contains (refPos)) {
							convexHull.Add (pointList [refPos]);
							alreadyProcessed.Add (refPos);
							hullIndex = refPos;
						}
						//   pointList.Remove(pointList[refPos])

					}

				}
			}
		}
		return convexHull;
	}
	
	/// <summary>
	/// Computes the intersection point of two lines.
	/// </summary>
	/// <param name="ps1">Start point of line 1</param>
	/// <param name="pe1">End point of line 1</param>
	/// <param name="ps2">Start point of line 2</param>
	/// <param name="pe2">End point of line 2</param>
	/// <returns>The intersection point</returns>
	/// <exception cref="Exception"></exception>
	private Vector2 lineIntersectionPoint(Vector2 ps1, Vector2 pe1, Vector2 ps2, Vector2 pe2){
		//Get A,B,C of first line - points : ps1 to pe1
		float A1 = pe1.y - ps1.y;
		float B1 = ps1.x - pe1.x;
		float C1 = A1 * ps1.x + B1 * ps1.y;

		//Get A,B,C of second line - points: ps2 to pe2
		float A2 = pe2.y - ps2.y;
		float B2 = ps2.x - pe2.x;
		float C2 = A2 * ps2.x + B2 * ps2.y;

		//Get delta and check if the lines are parallel
		float delta = A1 * B2 - A2 * B1;
		if (delta == 0) {
			throw new System.Exception ("Lines are parallel");
		}
		return new Vector2 ((B2 * C1 - B1 * C2) / delta, (A1 * C2 - A2 * C1) / delta);
	}
}
