﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;

/// <summary>
/// A viewpoint is set by a camera. Each viewpoint contains its own cuts and an occlusion Graph. A cut is precomputed 
/// based on the current target and the viewpoint.
/// </summary>
public class ViewPoint  {

	private GameObject camera;
	private List<Cut> cuts;
	private OcclusionGraph occlusionGraph;

	/// <summary>
	/// The constructor of the ViewPoint class
	/// </summary>
	/// <param name="camera">The camera of the ViewPoint</param>
	public ViewPoint(GameObject camera){
		cuts = new List<Cut> ();
		this.camera = camera;
		cuts = new List<Cut> ();
	}

	/// <summary>
	/// Adds a cut to the viewpoint
	/// </summary>
	/// <param name="cut">The cut to add to the ViewPoint</param>
	public void addCut(Cut cut){
		cuts.Add (cut);
	}

	/// <summary>
	/// Computes the occlusion graph for the viewpoint.
	/// </summary>
	public void computeOcclusionGraph(){
		occlusionGraph = new OcclusionGraph ();
		occlusionGraph.computeLayer (camera.GetComponent<Camera>());
		occlusionGraph.computeGraph (camera.GetComponent<Camera>());
	}
    
	/// <summary>
	/// Precomputes all cuts for the viewpoint with a specific target.
	/// </summary>
	/// <param name="target">The target</param>
	public void precomputeCuts(List<Transform> targets){
       
        GameObject[] allGameObj = GameObject.FindGameObjectsWithTag (Main.tagName);
		for (int i = 0; i < allGameObj.Length; i++) {
			allGameObj [i].SetActive (false);
		}
		for (int i = 0; i < cuts.Count; i++) {
			cuts [i].getTransform ().gameObject.SetActive (true);
			cuts [i].precompute (camera.transform, targets);
			cuts [i].getTransform ().gameObject.SetActive (false);
		}
		for (int i = 0; i < allGameObj.Length; i++) {
			allGameObj [i].SetActive (true);
		}
	}

	/// <summary>
	/// Performs the final cut to compute all cutting volumes.
	/// </summary>
	/// <param name="target">The target</param>
	public void performFinalCut(List<Transform> targets){
		//Debug.Log ("--------Perform final cut on target "+target.name+"--------");
		GameObject[] allObjects = GameObject.FindGameObjectsWithTag(Main.tagName);
		for (int i = 0; i < allObjects.Length; i++)
		{
			allObjects[i].SetActive(false);
		}
		foreach (Transform target in targets)
        {
            int occlusionLevel = 0;
            processOccluder(target, targets, occlusionLevel);
            for (int i = 0; i < allObjects.Length; i++)
            {
                allObjects[i].SetActive(true);
            }
        }
	}

	/// <summary>
	/// Resets the geometry object of the cuts
	/// </summary>
	public void resetGeometries()
	{
		foreach (var cut in cuts)
		{
			cut.resetGeometry();
		}
	}

	/// <summary>
	/// Computes the cut for all direct occluders of the parent Occluder. It is important that the target is always visible.
	/// A part of the occluder of the previous level should be visible as well.
	/// </summary>
	/// <param name="parentOccluder">The current occluder</param>
	/// <param name="target">The target geometry</param>
	/// <param name="occlusionLevel">The current level in the occlusion graph</param>
	public void processOccluder(Transform parentOccluder, List<Transform> targets, int occlusionLevel){
		Debug.Log ("Process Occluder " + parentOccluder.name);
		List<Transform> directOccluders = occlusionGraph.getDirectOccluders (parentOccluder.name);
		foreach (Transform occluder in directOccluders) {
			Cut cuttingVolume = cuts.First (c=> c.getTransform() == occluder);
            foreach(Transform target in targets)target.gameObject.SetActive(false);
            cuttingVolume.precompute(camera.transform, targets);
            foreach (Transform target in targets) target.gameObject.SetActive(true);
            cuttingVolume.computeFinalCut (targets, occlusionLevel, camera.GetComponent<Camera>());
			processOccluder (occluder, targets, occlusionLevel++);
		}
	}
	
	/// <summary>
	/// Sets the activation of the camera
	/// </summary>
	/// <param name="doActivate">Bool if the camera should be activated or not</param>
	public void activateCamera(bool doActivate){
		camera.SetActive (doActivate);
	}
}
